/*
 -- Multicore --
 !! FM synth in C !!
 and sequenced by a thread on core1
 
 8 bit PWM (500 kHz)
 */

#include "hardware/gpio.h"
#include "hardware/timer.h"
#include <hardware/irq.h>
#include <hardware/adc.h>
#include <hardware/pwm.h>
#include "hardware/interp.h"

#include "pico/stdlib.h"
#include "hardware/uart.h"
#include "stdio.h"
#include <string.h>
#include <pico/multicore.h>
#include <pico/divider.h>
#include "hardware/sync.h"
#include "stdlib.h"
#include "math.h"

// ==========================================
// === fixed point s19x12
// ==========================================
// s19x12 fixed point macros ==
typedef signed int fix;
#define mul(a,b) ((fix)(((( signed long long )(a))*(( signed long long )(b)))>>12)) //multiply two fixed 16:16
#define float_to_fix(a) ((fix)((a)*4096.0)) // 2^12
#define fix_to_float(a) ((float)(a)/4096.0)
#define fix_to_int(a)    ((int)((a)>>12))
#define int_to_fix(a)    ((fix)((a)<<12))
#define div(a,b) ((fix)((((signed long long)(a)<<12)/(b)))) 
#define absfix(a) abs(a)

// ==========================================
// === set up pwm
// ==========================================
uint pwm_slice_num,  pwm_chan_num ;
void pwm_setup(int pwm_pin){
  gpio_init(pwm_pin) ;
  gpio_set_function(pwm_pin, GPIO_FUNC_PWM) ;
  pwm_slice_num = pwm_gpio_to_slice_num(pwm_pin) ;
  // full speed clock => 256 counts in 2 uSec
  pwm_set_clkdiv(pwm_slice_num, 1.0f);
  pwm_set_clkdiv_mode(pwm_slice_num, PWM_DIV_FREE_RUNNING) ;
  // max count 256
  pwm_set_wrap(pwm_slice_num, 256);
  // initial for testing to half
  pwm_chan_num = pwm_gpio_to_channel(pwm_pin);
  pwm_set_chan_level(pwm_slice_num, pwm_chan_num , 128) ;
  pwm_set_enabled(pwm_slice_num, 1);
}

// 1/Fs in microseconds
volatile int alarm_period = 25 ;

// DDS variables 
unsigned int mod_inc[8], main_inc[8] ;
unsigned int current_mod_inc, current_main_inc ;
unsigned int mod_accum, main_accum ;

// amplitude paramters
fix max_mod_depth, current_mod_depth ; 
// waveform amplities
fix current_amp, max_amp = float_to_fix(120.0) ;

// timing in seconds
fix attack_time, mod_attack_time ;
fix decay_time, mod_decay_time, recip_decay_time ;
fix sustain_time, mod_sustain_time ;
fix onefix = int_to_fix(1);
// internal timing in samples
fix song_time, note_time ;
fix attack_inc, decay_inc, mod_attack_inc, mod_decay_inc, quad_decay_inc ;
float f_quad_decay_inc ;
// sine waves
fix sine_table[256] ;
fix mod_wave, main_wave ;
// inputs
float Fs, Fout, Fmod ;
//              C3     D3     E3     F3     G3     A3     B3     C4
float notes[8]={130.8, 146.8, 164.8, 174.6, 196.0, 220.0, 246.9, 261.6};
float a_time, d_time, mod_a_time, mod_d_time ; 
float s_time, mod_s_time, f_depth ;
//
int note_start=true ;
int linear_dk = 0 ;
int octave_num ;
 
// ==========================================
// === set up timer ISR NOT used in this pgm
// ==========================================
// === timer alarm ========================
// !! modifiying alarm zero trashes the cpu 
//        and causes LED  4 long - 4 short
// !! DO NOT USE alarm 0
/*
#define ALARM_NUM 1
#define ALARM_IRQ TIMER_IRQ_1
// ISR interval will be 10 uSec
//
// the actual ISR
static void alarm_irq(void) {
    // mark ISR entry
    //gpio_put(2,1);
    // Clear the alarm irq
    hw_clear_bits(&timer_hw->intr, 1u << ALARM_NUM);
    // arm the next interrupt
    // Write the lower 32 bits of the target time to the alarm to arm it
    timer_hw->alarm[ALARM_NUM] = timer_hw->timerawl + alarm_period ;
    //
    // === do DDS the usual way; ISR time 700 nS 
    //accum += increment ;
    //pwm_set_chan_level(pwm_slice_num, pwm_chan_num , sine_table[accum>>24]) ;

    // === interpolator version ; ISR time 600 nS
    // POP2 contents of pointer in result2 to pwm period 
    // Reading the 'pop' registers also clock the state of the interpolator
    pwm_set_chan_level(pwm_slice_num, pwm_chan_num , *(short*)(interp0->pop[2])) ;

    // mark ISR exit
    //gpio_put(2,0);
}
// set up the timer alarm ISR
static void alarm_in_us(uint32_t delay_us) {
    // Enable the interrupt for our alarm (the timer outputs 4 alarm irqs)
    hw_set_bits(&timer_hw->inte, 1u << ALARM_NUM);
    // Set irq handler for alarm irq
    irq_set_exclusive_handler(ALARM_IRQ, alarm_irq);
    // Enable the alarm irq
    irq_set_enabled(ALARM_IRQ, true);
    // Enable interrupt in block and at processor
    // Alarm is only 32 bits 
    uint64_t target = timer_hw->timerawl + delay_us;
    // Write the lower 32 bits of the target time to the alarm which
    // will arm it
    timer_hw->alarm[ALARM_NUM] = (uint32_t) target;   
}
*/

// ==========================================
// === protothreads globals
// ==========================================
// protothreads header
#include "pt_cornell_rp2040_v1_1_1.h"

// ==================================================
// === user's serial input thread
// ==================================================
// 
static PT_THREAD (protothread_serial(struct pt *pt))
{
  PT_BEGIN(pt);
    
    // convert alarm period in uSEc to rate
    Fs = 1.0/((float)alarm_period*1e-6) ;

    // data structure for interval timer
    PT_INTERVAL_INIT() ;
    //
    
    while(1) {
      
      // == Fout and Fmod are in Hz
      // == fm_depth is 0 to 10 or so
      // == times are in seconds

      // === first read
      sprintf(pt_serial_out_buffer, 
      "\n\r octave, attack, sustain, decay, linear?: ");
      serial_write ;
      serial_read ;
     // printf("return");
      // convert input string to number
      sscanf(pt_serial_in_buffer,"%d %f %f %f %d", 
        &octave_num, &a_time, &s_time, &d_time, &linear_dk) ;

      // === second read
      sprintf(pt_serial_out_buffer, 
      "\n\rFmod_ratio, max_fm_depth, attack, sustain, decay: ");
      serial_write ;
      serial_read ;
      // convert input string to number
      sscanf(pt_serial_in_buffer,"%f %f %f %f %f", 
        &Fmod, &f_depth, &mod_a_time, &mod_s_time, &mod_d_time) ;
        
      // conversion to intrnal units
      // increment = Fout/Fs * 2^32
      // octave number is based on a C3 to C4 table
      if (octave_num<1 || octave_num>6) octave_num = 3;
      float current_note ;
      for(int i=0; i<8; i++){
        current_note = notes[i] * pow(2, octave_num-3) ;
        main_inc[i] = current_note * pow(2,32 )/ Fs ;
        mod_inc[i] = Fmod * current_note * pow(2,32 )/ Fs ;
      }
      // fm modulation strength
      max_mod_depth = float_to_fix(f_depth * 100000) ;

      // convert main input times to sample number
      attack_time = float_to_fix(a_time * Fs) ;
      decay_time = float_to_fix(d_time * Fs) ;
      sustain_time = float_to_fix(s_time * Fs) ;
      // and now get increments
      attack_inc = div(max_amp, attack_time) ;
      // linear and parabolic fit
      decay_inc = div(max_amp, decay_time) ; 
      recip_decay_time = div(onefix, decay_time);
      //quad_decay_inc = mul((decay_inc<<1), recip_decay_time);
      // this need to be floating point becuase of squareing the decay time
      f_quad_decay_inc = (fix_to_float(decay_inc<<1)/fix_to_float(decay_time));

      // convert modulation input times to sample number
      mod_attack_time = float_to_fix(mod_a_time * Fs) ;
      mod_decay_time = float_to_fix(mod_d_time * Fs) ;
      mod_sustain_time = float_to_fix(mod_s_time * Fs) ;
      // and now get increments
      // precomputing increments means that only add/subtract is needed
      //mod_attack_inc = div(max_mod_depth, mod_attack_time) ;
     // mod_decay_inc = div(max_mod_depth, mod_decay_time) ;

     //
      
      //
      printf("start\n\r") ;
      // tell the synth thread to go
      linear_dk = 0 ;
      static int i, j ;
      static fix current_depth ;
      for(j=0; j<4; j++){
        current_depth = mul(int_to_fix(j),(max_mod_depth>>2));
        mod_attack_inc = div(current_depth, mod_attack_time) ;
        mod_decay_inc = div(current_depth, mod_decay_time) ;
        for(i=0; i<8; i++){
            current_main_inc = main_inc[i] ;
            current_mod_inc = mod_inc[i] ;
            note_start = true ;
            PT_YIELD_usec(20000) ;
            PT_YIELD_UNTIL(pt, current_amp<onefix);
            PT_YIELD_usec(10000) ;
        }
      }

      // NEVER exit while
    } // END WHILE(1)
  PT_END(pt);
} // timer thread

// ==================================================
// === toggle25 thread 
// ==================================================
// the on-board LED blinks
static PT_THREAD (protothread_toggle25(struct pt *pt))
{
  PT_BEGIN(pt);
    static bool LED_state = false ;
    
    // set up LED p25 to blink
    gpio_init(25) ;	
    gpio_set_dir(25, GPIO_OUT) ;
    gpio_put(25, true);
    // data structure for interval timer
    PT_INTERVAL_INIT() ;
     
    while(1) {
      // yield time 0.1 second
      PT_YIELD_INTERVAL(200000) ;

      // toggle the LED on PICO
      LED_state = LED_state? false : true ;
      gpio_put(25, LED_state);
      //
      // NEVER exit while
    } // END WHILE(1)
  PT_END(pt);
} // blink thread


// ==================================================
// === fast_ddds thread -- RUNNING on core 1
// ==================================================
// 
static PT_THREAD (protothread_fast_dds(struct pt *pt))
{
  PT_BEGIN(pt);
    
    // data structure for interval timer
    PT_INTERVAL_INIT() ;

    // start pwm on pin3
    #define pwm_dds 3
    pwm_setup(pwm_dds);

    // testing timing
    gpio_init(2) ;	
    gpio_set_dir(2, GPIO_OUT) ;
    gpio_put(2, true);

    while(1) {
      //
     // set rate to 1/alarm_period
     gpio_put(2, false);
     PT_YIELD_INTERVAL(alarm_period) ;
     gpio_put(2, true);

     // === using C only ===
     // start a burst on new data
      if(note_start) {
          // init the amplitude
          current_amp = attack_inc ;
          current_mod_depth = mod_attack_inc ;
          // reset the start flag
          note_start = false ;
          // reset envelope time
          note_time = 0 ;
          // phase lock the main frequency
          main_accum = 0 ;
      }
      // play the burst as long as the amplitude is positive
      // as it decays linearly
      if (current_amp > 0) {

        // update dds modulation freq
        mod_accum += current_mod_inc ;
        mod_wave = sine_table[mod_accum>>24] ;
        // get the instataneous modulation amp 
        // update modulation amplitude envelope 
        if (note_time < (mod_attack_time + mod_decay_time + mod_sustain_time)){
            current_mod_depth = (note_time <= mod_attack_time)? 
                current_mod_depth + mod_attack_inc : 
                (note_time <= mod_attack_time + mod_sustain_time)? current_mod_depth:
                    current_mod_depth - mod_decay_inc ;
        }
        else { 
            current_mod_depth = 0 ;
        }

        // set dds main freq and FM modulate it
        main_accum += current_main_inc + (unsigned int) mul(mod_wave, current_mod_depth) ;
        // update main waveform
        main_wave = sine_table[main_accum>>24] ;

        // get the instataneous amp 
        // update amplitude envelope 
        // linear EXCEPT for optional parabolic decay
        if (note_time < (attack_time + decay_time + sustain_time)){
            current_amp = (note_time <= attack_time)? 
                current_amp + attack_inc : 
                (note_time <= attack_time + sustain_time)? current_amp:
                    (linear_dk==1)? 
                      current_amp - (decay_inc) :
                      current_amp - (decay_inc<<1) + 
                        float_to_fix(f_quad_decay_inc * fix_to_float(note_time-attack_time-sustain_time)) ;
        }
        else {
            current_amp = 0 ;
        }
        // amplitide modulate and shift to the correct range for PWM
        main_wave = mul(main_wave, current_amp) ;
        // write final result  to 8-bit PWM    
        pwm_set_chan_level(pwm_slice_num, pwm_chan_num, fix_to_int(main_wave) + 128) ;
        //pwm_set_chan_level(pwm_slice_num, pwm_chan_num, fix_to_int(current_amp)) ;
        // move time ahead
        note_time += onefix ;
      }
      else {
          // set PWM to neutral level
        pwm_set_chan_level(pwm_slice_num, pwm_chan_num, 128) ;
      }
      // NEVER exit while
    } // END WHILE(1)
  PT_END(pt);
} // blink thread


// ========================================
// === core 1 main -- started in main below
// ========================================
void core1_main(){ 
  //  === add threads  ====================
  // for core 1
  pt_add_thread(protothread_fast_dds);
  //
  // === initalize the scheduler ==========
  pt_schedule_start ;
  // NEVER exits
  // ======================================
}

// ========================================
// === core 0 main
// ========================================
int main(){
  // set the clock
  //set_sys_clock_khz(250000, true); // 171us

  // dds table 10 bit values
  int i = 0 ;
  while (i<256) {
    // sine table is in naural +1/-1 range
    sine_table[i] = float_to_fix(sin(2*3.1416*i/256)) ;
    i++ ;
  }

  //start timer alarm IRQ
  // In THIS version, core1 takes the place of n ISR
  // toggle a pin for timing
  //gpio_init(2) ;	
  //gpio_set_dir(2, GPIO_OUT) ;
  //alarm_in_us(alarm_period);

  // start the serial i/o
  stdio_init_all() ;
  // announce the threader version on system reset
  printf("\n\rProtothreads RP2040 v1.11 two-core\n\r");
     
  // === add threads ======================
  // start core 1 threads
  multicore_reset_core1();
  multicore_launch_core1(&core1_main);

  // === config threads ===================
  // for core 0
  pt_add_thread(protothread_serial);
  pt_add_thread(protothread_toggle25);
  //
  // === initalize the scheduler ===============
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main
///////////
// end ////
///////////